"""Parse DS-PAW.log from relax task.

1. generate relaxing.xyz (extxyz format) showing the configurations during relaxation.
2. generate min_force.as showing the configuration with minimal force.
    2.1 convert `as` to `extxyz` format and `cif` format for checking
3. collect max force and energy for each relaxation ionic step and print them as table
    3.1 save to csv file

Note that turn on vdw will make DS-PAW.log content format changed, thus we need two separate functions.
- without vdw, DS-PAW.log content looks like the following:
...
## STRUCTURE  ##
Lattice vectors (Angstrom)
  0.000000   2.750000   2.750000
  2.750000   0.000000   2.750000
  2.750000   2.750000   0.000000
CoordinateType                   = Direct
Atom position:
Si   0.885000   0.875000   0.875000
...

## KPOINTS ##
Reduced kpoints size             = 282
Kpoints reciprocal coordinates and weight:
  0.000000  0.000000  0.000000  1.000000
  0.100000 -0.000000  0.000000  2.000000
  0.200000  0.000000  0.000000  2.000000
...
  0.500000  0.500000  0.500000  1.000000

## PARAMETERS ##
task = relax
...

## INITIALIZATION ##
Load structure from "structure.as" successfully!
------------------------------------------------------------
LOOP 1:
------------------------------------------------------------
#  iter |      Etot(eV)       dE(eV)         time
#   1   |    -186.196437  -1.861964e+02     0.151 s
#   2   |    -210.417628  -2.422119e+01     0.134 s
...
#  17   |    -216.967842  -1.796438e-05     0.173 s

------------------------------------------------------------
## STRUCTURE ##
Lattice vectors (Angstrom)
  0.000000   2.750000   2.750000
  2.750000   0.000000   2.750000
  2.750000   2.750000   0.000000
CoordinateType                   = Direct
Atom Position and force:
Si   0.885000   0.875000   0.875000  -0.024045  -0.345558  -0.345558
...

------------------------------------------------------------
Total force(eV/Angstrom):    0.001701   0.000124   0.000125
Max force:    0.489284 eV/Angstrom at atom index: 1
------------------------------------------------------------
Calculating force and stress:          0.155 s
Total calculation time      :          2.986 s

------------------------------------------------------------
## RELAX ##
Pressure              P  =            -0.38 kbar
Volume                V  =            41.59 Ang**3
------------------------------------------------------------
LOOP 2:
------------------------------------------------------------
#  iter |      Etot(eV)       dE(eV)         time
#   1   |    -216.976553  -8.728839e-03     0.205 s
...
#   3   |    -216.976886  -2.038071e-05     0.185 s

------------------------------------------------------------
## STRUCTURE ##
Lattice vectors (Angstrom)
  0.000000   2.750000   2.750000
  2.750000   0.000000   2.750000
  2.750000   2.750000   0.000000
CoordinateType                   = Direct
Atom Position and force:
Si   0.879149   0.874789   0.874789   0.028458   0.074841   0.074841
...

------------------------------------------------------------
Total force(eV/Angstrom):   -0.000703   0.000766   0.000766
Max force:    0.109600 eV/Angstrom at atom index: 1
------------------------------------------------------------
Calculating force and stress:          0.154 s
Total calculation time      :          0.789 s

------------------------------------------------------------
## RELAX ##
Pressure              P  =            -0.39 kbar
Volume                V  =            41.59 Ang**3
------------------------------------------------------------
...(Other Loops)
Convergence criteria (relax.convergence) reached on loop 3.
Total relaxation time       :           4.315 s
...(some tail info)


- with vdw, content looks like:
...(header is same as above)
## INITIALIZATION ##
Load structure from "structure.as" successfully!
------------------------------------------------------------
LOOP 1:
------------------------------------------------------------
#  iter |      Etot(eV)       dE(eV)         time
#   1   |    -512.851127  -5.128511e+02     1.000 s
...
#  11   |    -625.480265  -4.310149e-06     0.423 s

## VDW CORRECTION ##
Functional       :       pbe
Damping function paremeters:
rs6 =   1.000000
s18 =   0.722000
Correction Energy:        -0.306427 eV
Total Energy     :       -625.786692 eV
------------------------------------------------------------
## STRUCTURE ##
Lattice vectors (Angstrom)
  2.463267   0.000000  -0.000000
 -1.231633   2.133252   0.000000
  0.000000   0.000000   7.766420
CoordinateType                   = Direct
Atom Position and force:
 C   0.000000   0.000000   0.250000   0.000000   0.000000  -0.000000
...

------------------------------------------------------------
Total force(eV/Angstrom):   -0.000126   0.000148  -0.000000
Max force:    0.000000 eV/Angstrom at atom index: 4
------------------------------------------------------------
Calculating force and stress:          0.629 s
Total calculation time      :          6.812 s

------------------------------------------------------------
## RELAX ##
Pressure              P  =            -0.22 kbar
Volume                V  =            40.81 Ang**3
------------------------------------------------------------
LOOP 2:
------------------------------------------------------------
#  iter |      Etot(eV)       dE(eV)         time
#   1   |    -625.482228  -1.967494e-03     0.449 s
...
#   4   |    -625.479282  -6.023662e-06     0.448 s

## VDW CORRECTION ##
Functional       :       pbe
Damping function paremeters:
rs6 =   1.000000
s18 =   0.722000
Correction Energy:        -0.308962 eV
Total Energy     :       -625.788244 eV
------------------------------------------------------------
## STRUCTURE ##
Lattice vectors (Angstrom)
  2.458111   0.000000  -0.000000
 -1.229056   2.128787   0.000000
  0.000000   0.000000   7.719465
CoordinateType                   = Direct
Atom Position and force:
 C   0.000000   0.000000   0.250000   0.000000   0.000000   0.000000
。。。

------------------------------------------------------------
Total force(eV/Angstrom):   -0.000125   0.000147  -0.000000
Max force:    0.000000 eV/Angstrom at atom index: 4
------------------------------------------------------------
Calculating force and stress:          0.463 s
Total calculation time      :          2.342 s

------------------------------------------------------------
## RELAX ##
Pressure              P  =             0.08 kbar
Volume                V  =            40.39 Ang**3
------------------------------------------------------------
...(Other Loops)
Convergence criteria (relax.convergence) reached on loop 13.
Total relaxation time       :          40.346 s
...(some tail info)
"""


def _get_last_relax_blocks(logfile: str) -> list[str]:
    """Return all lines of the last relax (starting from the last LOOP 1:)"""
    with open(logfile, encoding="utf-8") as f:
        lines = f.readlines()
    # Use simple string matching, avoid regular expressions
    loop1_idx = []
    for i in range(len(lines) - 1):
        if lines[i].strip("- \n") == "" and lines[i + 1].strip().startswith("LOOP 1:"):
            loop1_idx.append(i)
    if not loop1_idx:
        raise RuntimeError("No LOOP 1 found in log")
    start = loop1_idx[-1]
    return lines[start:]


def _split_loops(lines: list[str]) -> list[list[str]]:
    """Split by LOOP"""
    loop_starts = []
    for i in range(len(lines) - 1):
        if lines[i].strip("- \n") == "" and lines[i + 1].strip().startswith("LOOP"):
            loop_starts.append(i)
    blocks = []
    for idx, s in enumerate(loop_starts):
        e = loop_starts[idx + 1] if idx + 1 < len(loop_starts) else len(lines)
        blocks.append(lines[s:e])
    return blocks


def _parse_structure_block(block: list[str]):
    """Parse the structure block and return a list of elements, fractional coordinates, forces, and lattice"""
    # Find lattice vectors
    lat_idx = None
    for i, line in enumerate(block):
        if "Lattice vectors" in line:
            lat_idx = i
            break
    if lat_idx is None:
        return None, None, None, None
    lattice = []
    for j in range(1, 4):
        lattice.append([float(x) for x in block[lat_idx + j].split()])
    # Find Atom Position and force:
    atom_start = None
    for idx in range(lat_idx + 4, len(block)):
        if "Atom Position and force" in block[idx]:
            atom_start = idx + 1
            break
    if atom_start is None:
        return None, None, None, None
    atoms: list[str] = []
    coords: list[list[float]] = []
    forces: list[list[float]] = []
    for line in block[atom_start:]:
        if not line.strip() or line.startswith("-"):
            break
        parts = line.split()
        atoms.append(parts[0])
        coords.append([float(x) for x in parts[1:4]])
        forces.append([float(x) for x in parts[4:7]])
    return atoms, coords, forces, lattice


def _parse_energy_block(block: list[str], vdw: bool = False):
    """Parse energy, take Total Energy in vdw mode, otherwise take Etot of the last iteration"""
    if vdw:
        # Find "Total Energy" in reverse order
        for line in reversed(block):
            if "Total Energy" in line:
                try:
                    return float(line.split(":")[-1].split()[0])
                except Exception:
                    continue
    else:
        # Find the last line starting with # that contains 'iter'
        etot = None
        for line in block:
            if line.strip().startswith("#") and "|" in line:
                parts = line.split("|")
                if len(parts) > 1:
                    try:
                        etot = float(parts[1].strip().split()[0])
                    except Exception:
                        continue
        return etot
    return None


def _parse_max_force_block(block: list[str]):
    """Parsing Max force"""
    for line in block:
        if line.strip().startswith("Max force:"):
            try:
                return float(line.split()[2])
            except Exception:
                continue
    return None


def _infer_selective_dynamics(
    forces: list[list[float]], tol: float = 1e-8
) -> list[list[str]]:
    """Determine the degrees of freedom for each atom based on forces (0 for T, non-zero for F)"""
    sd = []
    for f in forces:
        sd.append([("T" if abs(x) < tol else "F") for x in f])
    return sd


def get_relaxing_xyz(logfile: str = "DS-PAW.log"):
    """Generate relaxing.xyz (extxyz format), record the structure at each step"""
    from dspawpy.io.write import _write_xyz_traj

    lines = _get_last_relax_blocks(logfile)
    loops = _split_loops(lines)
    structures = []
    for loop in loops:
        atoms, coords, forces, lattice = _parse_structure_block(loop)
        if atoms is None or lattice is None or coords is None or forces is None:
            continue
        from pymatgen.core import Lattice, Structure

        struct = Structure(
            Lattice(lattice),
            atoms,
            coords,
            coords_are_cartesian=False,
        )
        structures.append(struct)
    if structures:
        _write_xyz_traj(structures, xyzfile="relaxing.xyz")
        print("==> relaxing.xyz")
    else:
        print("No structure found.")


def get_min_structure(logfile: str = "DS-PAW.log"):
    """Generate min_force.as, the structure corresponding to the minimum max force, retaining degree of freedom information"""
    lines = _get_last_relax_blocks(logfile)
    loops = _split_loops(lines)
    minf = float("inf")
    minidx = -1
    max_forces = []
    for i, loop in enumerate(loops):
        f = _parse_max_force_block(loop)
        max_forces.append(f)
        if f is not None and f < minf:
            minf = f
            minidx = i
    if minidx == -1:
        print("No force found.")
        return
    atoms, coords, forces, lattice = _parse_structure_block(loops[minidx])
    if atoms is None or lattice is None or coords is None or forces is None:
        print("Can't determine min_force structure")
        exit(1)
    from pymatgen.core import Lattice, Structure

    struct = Structure(
        Lattice(lattice),
        atoms,
        coords,
        coords_are_cartesian=False,
    )
    # Write AS file with degrees of freedom and header
    sd = _infer_selective_dynamics(forces)
    with open("min_force.as", "w") as f:
        f.write("Total number of atoms\n")
        f.write(f"{len(atoms)}\n")
        f.write("Lattice\n")
        for v in lattice:
            f.write("  " + "  ".join(f"{x:.8f}" for x in v) + "\n")
        f.write("Direct Fix\n")
        for i in range(len(atoms)):
            f.write(
                f"{atoms[i]}  "
                + "  ".join(f"{x:.16f}" for x in coords[i])
                + "    "
                + "  ".join(sd[i])
                + "\n"
            )
    print("==> min_force.as")
    # Optional: convert to extxyz/cif
    from dspawpy.io.write import _write_xyz_traj

    _write_xyz_traj([struct], xyzfile="min_force.xyz")
    try:
        from pymatgen.io.cif import CifWriter

        CifWriter(struct).write_file("min_force.cif")
    except Exception:
        pass


def get_max_forces(logfile: str = "DS-PAW.log"):
    """Output the maximum force for each step"""
    lines = _get_last_relax_blocks(logfile)
    loops = _split_loops(lines)
    max_forces = []
    for loop in loops:
        f = _parse_max_force_block(loop)
        max_forces.append(f)

    return max_forces


def get_energies(logfile: str = "DS-PAW.log"):
    """Output the energy at each step"""
    lines = _get_last_relax_blocks(logfile)
    loops = _split_loops(lines)
    energies = []
    for loop in loops:
        e = _parse_energy_block(loop, vdw=False)
        energies.append(e)

    return energies


def get_max_forces_and_energies(logfile: str = "DS-PAW.log", relative: bool = False):
    """Output the maximum forces and energies for each step and save them as a CSV file"""
    max_forces = get_max_forces(logfile)
    energies = get_energies(logfile)
    if relative:
        max_forces = [f - max_forces[0] for f in max_forces]
        energies = [e - energies[0] for e in energies]
    import polars as pl

    df = pl.DataFrame(
        data={
            "step": list(range(len(max_forces))),
            "force (eV/Angstrom)": max_forces,
            "energy (eV)": energies,
        }
    )
    print(df)
    df.write_csv("relax_summary.csv")
    print("==> relax_summary.csv")


# Implementation for VDW mode
def get_relaxing_xyz_vdw(logfile: str = "DS-PAW.log"):
    lines = _get_last_relax_blocks(logfile)
    loops = _split_loops(lines)
    structures = []
    for loop in loops:
        atoms, coords, forces, lattice = _parse_structure_block(loop)
        if atoms is None or lattice is None or coords is None or forces is None:
            continue
        from pymatgen.core import Lattice, Structure

        struct = Structure(
            Lattice(lattice),
            atoms,
            coords,
            coords_are_cartesian=False,
        )
        struct.add_site_property("forces", forces)
        structures.append(struct)
    if structures:
        from dspawpy.io.write import _write_xyz_traj

        _write_xyz_traj(structures, xyzfile="relaxing_vdw.xyz")
        print("==> relaxing_vdw.xyz")
    else:
        print("No structure found.")


def get_min_structure_vdw(logfile: str = "DS-PAW.log"):
    """Generate min_force_vdw.as in vdw mode, preserving the degree of freedom information"""
    lines = _get_last_relax_blocks(logfile)
    loops = _split_loops(lines)
    minf = float("inf")
    minidx = -1
    for i, loop in enumerate(loops):
        f = _parse_max_force_block(loop)
        if f is not None and f < minf:
            minf = f
            minidx = i
    if minidx == -1:
        print("No force found.")
        return
    atoms, coords, forces, lattice = _parse_structure_block(loops[minidx])
    if atoms is None or lattice is None or coords is None or forces is None:
        print("Can't determine min_force structure")
        exit(1)
    from pymatgen.core import Lattice, Structure

    struct = Structure(
        Lattice(lattice),
        atoms,
        coords,
        coords_are_cartesian=False,
    )
    sd = _infer_selective_dynamics(forces)
    with open("min_force_vdw.as", "w") as f:
        f.write("Total number of atoms\n")
        f.write(f"{len(atoms)}\n")
        f.write("Lattice\n")
        for v in lattice:
            f.write("  " + "  ".join(f"{x:.8f}" for x in v) + "\n")
        f.write("Direct Fix\n")
        for i in range(len(atoms)):
            f.write(
                f"{atoms[i]}  "
                + "  ".join(f"{x:.16f}" for x in coords[i])
                + "    "
                + "  ".join(sd[i])
                + "\n"
            )
    from dspawpy.io.write import _write_xyz_traj

    _write_xyz_traj([struct], xyzfile="min_force_vdw.xyz")
    try:
        from pymatgen.io.cif import CifWriter

        CifWriter(struct).write_file("min_force_vdw.cif")
    except Exception:
        pass
    print("==> min_force_vdw.as")


def get_max_forces_vdw(logfile: str = "DS-PAW.log"):
    lines = _get_last_relax_blocks(logfile)
    loops = _split_loops(lines)
    max_forces = []
    for loop in loops:
        f = _parse_max_force_block(loop)
        max_forces.append(f)

    return max_forces


def get_energies_vdw(logfile: str = "DS-PAW.log"):
    lines = _get_last_relax_blocks(logfile)
    loops = _split_loops(lines)
    energies = []
    for loop in loops:
        e = _parse_energy_block(loop, vdw=True)
        energies.append(e)

    return energies


def get_max_forces_and_energies_vdw(
    logfile: str = "DS-PAW.log", relative: bool = False
):
    max_forces = get_max_forces_vdw(logfile)
    energies = get_energies_vdw(logfile)

    if relative:
        max_forces = [f - max_forces[0] for f in max_forces]
        energies = [e - energies[0] for e in energies]

    import polars as pl

    df = pl.DataFrame(
        data={
            "step": list(range(len(max_forces))),
            "force (eV/Angstrom)": max_forces,
            "energy (eV)": energies,
        }
    )
    print(df)
    df.write_csv("relax_summary_vdw.csv")

    print("==> relax_summary_vdw.csv")


def read_relax_data(
    logfile: str = "DS-PAW.log",
    print_efs: bool = True,
    relative: bool = False,
    write_traj: bool = True,
    write_as: bool = True,
):
    """Read relax data from log file, not h5/json, because they lack data.

    Call this function will print info to screen directly.

    Parameters
    ----------
    logfile:
        location of "DS-PAW.log", such as '../another/DS-PAW.log'.
    print_efs:
        whether to print energies and forces table.
    relative:
        whether to print energies and forces relative to the first step value.
    write_traj:
        whether to write traj xyz file
    write_as:
        whether to write structure.as file corresponding to the minimal force.

    Examples
    --------
    >>> from dspawpy.analysis.post_relax import read_relax_data
    >>> read_relax_data(logfile='tests/2.1/DS-PAW.log') # doctest: +ELLIPSIS
    ==> ...relaxing...xyz
    ==> relaxing.xyz
    ==> min_force.as
    ==> ...min_force...xyz
    shape: (3, 3)
    ┌──────┬─────────────────────┬─────────────┐
    │ step ┆ force (eV/Angstrom) ┆ energy (eV) │
    ╞══════╪═════════════════════╪═════════════╡
    │ 0    ┆ 0.489284            ┆ -216.967842 │
    │ 1    ┆ 0.1096              ┆ -216.976886 │
    │ 2    ┆ 0.024697            ┆ -216.977327 │
    └──────┴─────────────────────┴─────────────┘
    ==> relax_summary.csv

    >>> read_relax_data(logfile='tests/2.1/DS-PAW.log', relative=True, write_traj=False, write_as=False)
    shape: (3, 3)
    ┌──────┬─────────────────────┬─────────────┐
    │ step ┆ force (eV/Angstrom) ┆ energy (eV) │
    ╞══════╪═════════════════════╪═════════════╡
    │ 0    ┆ 0.0                 ┆ 0.0         │
    │ 1    ┆ -0.379684           ┆ -0.009044   │
    │ 2    ┆ -0.464587           ┆ -0.009485   │
    └──────┴─────────────────────┴─────────────┘
    ==> relax_summary.csv
    """
    with open(logfile, encoding="utf-8") as f:
        if "## VDW CORRECTION ##" in f.read():
            if write_traj:
                get_relaxing_xyz_vdw(logfile)
            if write_as:
                get_min_structure_vdw(logfile)
            if print_efs:
                get_max_forces_and_energies_vdw(logfile, relative)

        else:
            if write_traj:
                get_relaxing_xyz(logfile)
            if write_as:
                get_min_structure(logfile)
            if print_efs:
                get_max_forces_and_energies(logfile, relative)



import os
import shutil
from typing import TYPE_CHECKING, Optional, Sequence, Union

from loguru import logger

if TYPE_CHECKING:
    import numpy as np

Bohr = 0.52917721067  # Angstrom


def handle_duplicated_output(dst: str) -> str:
    if os.path.exists(dst):
        from datetime import datetime

        ow_mode = os.getenv("OVERWRITE", "default")

        if ow_mode == "yes":
            logger.warning(
                f"Overwriting existing {dst} because OVERWRITE = yes",
                category=UserWarning,
            )
            # file does not require manual deletion
            if os.path.isdir(dst):
                shutil.rmtree(dst)
        elif ow_mode == "no":
            logger.warning(
                "Refuse to write because OVERWRITE = off",
                category=UserWarning,
            )
            raise FileExistsError(f"{dst} already exists")
        elif ow_mode == "bk":
            # 保留文件扩展名
            base, ext = os.path.splitext(dst)
            new_location = f"{base}_" + datetime.now().strftime("%Y%m%d%H%M%S") + ext
            logger.warning(
                f"Renaming existing {dst} to {new_location} because OVERWRITE = bk",
                category=UserWarning,
            )
            shutil.move(
                dst,
                new_location,
            )
        else:  # default
            orig_dst = dst
            # 保留文件扩展名
            base, ext = os.path.splitext(dst)
            dst = base + datetime.now().strftime("%Y%m%d%H%M%S") + ext
            logger.warning(
                f"Put {orig_dst} to {dst} to avoid overwriting existing same-name file because OVERWRITE = {ow_mode}",
                category=UserWarning,
            )

    return dst


def _write_xyz_traj(
    structures: Sequence,
    xyzfile: str = "aimdTraj.xyz",
):
    r"""Save the trajectory file in XYZ format

    Parameters
    ----------
    structures:
        List of pymatgen Structures
    xyzfile:
        Write the trajectory file in xyz format, default is aimdTraj.xyz

    """

    if not isinstance(structures, Sequence):  # single Structure
        structures = [structures]
    absxyz = handle_duplicated_output(os.path.abspath(xyzfile))
    os.makedirs(os.path.dirname(absxyz), exist_ok=True)
    with open(absxyz, "w") as f:
        # Nstep
        for _, structure in enumerate(structures):
            elements = [s.species_string for s in structure.sites]
            f.write("%d\n" % len(elements))
            # lattice
            lm = structure.lattice.matrix
            f.write(
                'Lattice="%f %f %f %f %f %f %f %f %f" Properties=species:S:1:pos:R:3 pbc="T T T"\n'
                % (
                    lm[0, 0],
                    lm[0, 1],
                    lm[0, 2],
                    lm[1, 0],
                    lm[1, 1],
                    lm[1, 2],
                    lm[2, 0],
                    lm[2, 1],
                    lm[2, 2],
                ),
            )
            # position and element
            poses = structure.cart_coords
            for j in range(len(elements)):
                f.write(
                    "%s %f %f %f\n"
                    % (elements[j], poses[j, 0], poses[j, 1], poses[j, 2]),
                )

    print(f"==> {absxyz}")


def _write_dump_traj(
    structures: Sequence,
    dumpfile: str = "aimdTraj.dump",
):
    r"""Save trajectory file in LAMMPS dump format, currently only supports orthogonal cells

    Parameters
    ----------
    structures:
        A list of pymatgen Structures
    dumpfile:
        The dump format trajectory file name, default to aimdTraj.dump

    """

    if not isinstance(structures, Sequence):  # single Structure
        structures = [structures]
    absdump = handle_duplicated_output(os.path.abspath(dumpfile))
    os.makedirs(os.path.dirname(absdump), exist_ok=True)
    from dspawpy.io.read import _get_lammps_non_orthogonal_box

    with open(absdump, "w") as f:
        for n, structure in enumerate(structures):
            lat = structure.lattice.matrix
            elements = [s.species_string for s in structure.sites]
            poses = structure.cart_coords

            box_bounds = _get_lammps_non_orthogonal_box(lat)
            f.write("ITEM: TIMESTEP\n%d\n" % n)
            f.write("ITEM: NUMBER OF ATOMS\n%d\n" % (len(elements)))
            f.write("ITEM: BOX BOUNDS xy xz yz xx yy zz\n")
            f.write(
                "%f %f %f\n%f %f %f\n %f %f %f\n"
                % (
                    box_bounds[0][0],
                    box_bounds[0][1],
                    box_bounds[0][2],
                    box_bounds[1][0],
                    box_bounds[1][1],
                    box_bounds[1][2],
                    box_bounds[2][0],
                    box_bounds[2][1],
                    box_bounds[2][2],
                ),
            )
            f.write("ITEM: ATOMS type x y z id\n")
            for i in range(len(elements)):
                f.write(
                    "%s %f %f %f %d\n"
                    % (
                        elements[i],
                        poses[i, 0],
                        poses[i, 1],
                        poses[i, 2],
                        i + 1,
                    ),
                )
    print(f"==> {absdump}")


def write_VESTA(
    in_filename: str,
    data_type: str,
    out_filename: str = "DS-PAW.cube",
    subtype: Optional[str] = None,
    format: Optional[str] = "cube",
    compact: bool = False,
    inorm: bool = False,
    gridsize: Optional[Sequence[int]] = None,
):
    """Read data from a json or h5 file containing electronic system information and write to a VESTA formatted file.

    Parameters
    ----------
    in_filename:
        Path to a json or h5 file containing electronic system information
    data_type:
        Data type, supported values are "rho", "potential", "elf", "pcharge", "rhoBound"
    out_filename:
        Output file path, default "DS-PAW.cube"
    subtype:
        Used to specify the subtype of data_type, default is None, which will read the TotalElectrostaticPotential data of potential
    format:
        Output data format, supports "cube" and "vesta" ("vasp"), default is "cube", case-insensitive
    compact:
        Each data point for each grid is placed on a new line, reducing the file size by decreasing the number of spaces (this does not affect the parsing of VESTA software), default is False
    inorm:
        Whether to normalize the volume data so that the sum is 1, default is False
    gridsize:
        The redefined number of grid points, in the format (ngx, ngy, ngz), default is None, which uses the original number of grid points

    Returns
    -------
    out_filename:
        VESTA formatted file

    Examples
    --------
    >>> from dspawpy.io.write import write_VESTA
    >>> write_VESTA("tests/2.2/rho.json", "rho", out_filename='tests/outputs/doctest/rho.cube') # doctest: +ELLIPSIS
    ==> ...rho...cube

    >>> from dspawpy.io.write import write_VESTA
    >>> write_VESTA(
    ...     in_filename="tests/2.7/potential.h5",
    ...     data_type="potential",
    ...     out_filename="tests/outputs/doctest/my_potential.cube",
    ...     subtype='TotalElectrostaticPotential', # or 'TotalLocalPotential'
    ...     gridsize=(50,50,50), # all integer, can be larger or less than the original gridsize
    ... ) # doctest: +ELLIPSIS
    Interpolating volumetric data...
    volumetric data interpolated
    ==> ...my_potential...cube
    >>> write_VESTA(
    ...     in_filename="tests/2.8/elf.h5",
    ...     data_type="elf",
    ...     out_filename="tests/outputs/doctest/elf.cube",
    ... ) # doctest: +ELLIPSIS
    ==> ...elf...cube
    >>> write_VESTA(
    ...     in_filename="tests/2.9/pcharge.h5",
    ...     data_type="pcharge",
    ...     out_filename="tests/outputs/doctest/pcharge.cube",
    ... ) # doctest: +ELLIPSIS
    ==> ...pcharge...cube
    >>> write_VESTA(
    ...     in_filename="tests/2.7/potential.h5",
    ...     data_type="potential",
    ...     out_filename="tests/outputs/doctest/my_potential.vasp",
    ...     subtype='TotalElectrostaticPotential', # or 'TotalLocalPotential'
    ...     gridsize=(50,50,50), # all integer, can be larger or less than the original gridsize
    ... ) # doctest: +ELLIPSIS
    Interpolating volumetric data...
    volumetric data interpolated
    ==> ...my_potential...vasp
    >>> write_VESTA(
    ...     in_filename="tests/2.8/elf.h5",
    ...     data_type="elf",
    ...     out_filename="tests/outputs/doctest/elf.vasp",
    ... ) # doctest: +ELLIPSIS
    ==> ...elf...vasp
    >>> write_VESTA(
    ...     in_filename="tests/2.9/pcharge.h5",
    ...     data_type="pcharge",
    ...     out_filename="tests/outputs/doctest/pcharge.vasp",
    ... ) # doctest: +ELLIPSIS
    ==> ...pcharge...vasp

    >>> write_VESTA(
    ...     in_filename="tests/2.7/potential.h5",
    ...     data_type="potential",
    ...     out_filename="tests/outputs/doctest/my_potential.txt",
    ...     subtype='TotalElectrostaticPotential', # or 'TotalLocalPotential'
    ...     gridsize=(50,50,50), # all integer, can be larger or less than the original gridsize
    ... ) # doctest: +ELLIPSIS
    Interpolating volumetric data...
    volumetric data interpolated
    ==> ...my_potential...txt
    >>> with open("tests/outputs/doctest/my_potential.txt") as t:
    ...     contents = t.readlines()
    ...     for line in contents[:10]:
    ...         print(line.strip())
    # 2 atoms
    # 50 50 50 grid size
    # x y z value
    0.000 0.000 0.000      0.3279418
    0.055 0.055 0.000     -0.0740864
    0.110 0.110 0.000     -0.8811763
    0.165 0.165 0.000     -2.1283865
    0.220 0.220 0.000     -4.0559145
    0.275 0.275 0.000     -6.8291030
    0.330 0.330 0.000    -10.1550909

    """
    structure, grid, vd = _get_structure_grid_vd(in_filename, data_type, subtype)

    _write_specific_format(
        structure,
        grid,
        vd,
        out_filename,
        gridsize,
        format,
        compact,
        inorm,
    )


def _get_structure_grid_vd(
    in_filename: str,
    data_type: str,
    subtype: Optional[str] = None,
) -> tuple:
    import numpy as np

    from dspawpy.io.structure import read

    if in_filename.endswith(".h5"):
        from dspawpy.io.read import load_h5

        data = load_h5(in_filename)
        structure = read(in_filename)[0]
        grid = np.asarray(data["/AtomInfo/Grid"])
        if data_type == "rho" or data_type == "rhoBound":
            vd_array = np.asarray(data["/Rho/TotalCharge"])
        elif data_type == "pcharge":
            vd_array = np.asarray(data["/Pcharge/1/TotalCharge"])
        elif data_type == "potential":
            if subtype is None:
                subtype = "TotalElectrostaticPotential"
            vd_array = np.asarray(data[f"/Potential/{subtype}"])
        elif data_type == "elf":  # has to flatten and reshape to keep order
            vd_array = np.asarray(data["/ELF/TotalELF"]).flatten()
        else:
            raise NotImplementedError(
                f"{data_type}, {subtype}, Only support rho/potential/elf/pcharge/rhoBound",
            )

    elif in_filename.endswith(".json"):
        with open(in_filename) as fin:
            from json import load

            data = load(fin)
        structure = read(in_filename)[0]
        grid = np.asarray(data["AtomInfo"]["Grid"])  # get grid array
        if data_type == "rho" or data_type == "rhoBound":
            vd_array = np.asarray(data["Rho"]["TotalCharge"])
        elif data_type == "pcharge":
            vd_array = np.asarray(data["Pcharge"][0]["TotalCharge"])
        elif data_type == "potential":
            if subtype is None:
                subtype = "TotalElectrostaticPotential"
            vd_array = np.asarray(data["Potential"][subtype])
        elif data_type == "elf":
            vd_array = np.asarray(data["ELF"]["TotalELF"])
        else:
            raise NotImplementedError(
                f"{data_type}, {subtype}, Only support rho/potential/elf/pcharge/rhoBound",
            )

    else:
        raise NotImplementedError("Only support json/h5 format")

    vd = vd_array.reshape(grid, order="F")

    return structure, grid, vd


def write_delta_rho_vesta(
    total: str,
    individuals: list[str],
    output: str = "delta_rho.cube",
    format: str = "cube",
    compact: bool = False,
    inorm: bool = False,
    gridsize: Optional[Sequence] = None,
    data_type: Optional[str] = "rho",
    subtype: Optional[str] = None,
):
    """Charge density differential visualization

    DeviceStudio does not currently support large files; it is temporarily written in a format that can be opened with VESTA.

    Parameters
    ----------
    total:
        Path to the total charge density file of the system, can be in h5 or json format
    individuals:
        Paths to the charge density files of each component in the system, can be in h5 or json format
    output:
        Output file path, default "delta_rho.cube"
    format:
        Output data format, supports "cube" and "vasp", default to "cube"
    compact:
        Each data point for each grid is placed on a new line, and the file size is reduced by reducing the number of spaces (this does not affect the parsing by VESTA software), default is False
    inorm:
        Whether to normalize the volume data so that the sum is 1, default is False
    gridsize:
        Redefined grid number, format as (ngx, ngy, ngz), default is None, use the original grid number

    Returns
    -------
    output:
        A charge density file after the difference of charges (total - individual1 - individual2 - ...)

    Examples
    --------
    >>> from dspawpy.io.write import write_delta_rho_vesta
    >>> write_delta_rho_vesta(total='tests/supplement/AB.h5',
    ...     individuals=['tests/supplement/A.h5', 'tests/supplement/B.h5'],
    ...     output='tests/outputs/doctest/delta_rho.cube') # doctest: +ELLIPSIS
    ==> ...delta_rho...cube

    """
    import numpy as np

    files = [total] + individuals
    data_type = data_type or "rho"

    ss = []
    grids = []
    vds = []
    for f in files:
        structure, grid, vd = _get_structure_grid_vd(f, data_type, subtype)
        ss.append(structure)
        grids.append(grid)
        vds.append(vd)

    if not all(np.allclose(grids[0], grids[i + 1]) for i in range(len(individuals))):
        raise ValueError(f"grids not consistent: {grids}")
    if any(ss[0] == ss[i + 1] for i in range(len(individuals))):
        raise ValueError("number of structure is less than 3")

    _vd = vds[0]
    for i in range(len(individuals)):
        _vd -= vds[i + 1]

    _write_specific_format(
        ss[0],
        grids[0],
        _vd,
        output,
        gridsize,
        format=format,
        compact=compact,
        inorm=inorm,
    )


def to_file(
    structure,
    filename: str,
    fmt: Optional[str] = None,
    coords_are_cartesian: bool = True,
):
    r"""Deprecated. Use :func:`dspawpy.io.structure.write` instead."""
    logger.warning(
        "dspawpy.io.write.to_file is deprecatedUse dspawpy.io.structure.write instead.",
        DeprecationWarning,
    )
    from dspawpy.io.structure import write

    write(structure, filename, fmt, coords_are_cartesian)


def _write_atoms(fileobj, structure, idirect: bool = False):
    fileobj.write("DS-PAW Structure\n")
    fileobj.write("  1.00\n")
    lattice = structure.lattice.matrix.reshape(-1, 1)
    fileobj.write(
        "%20.14f %20.14f %20.14f\n" % (lattice[0][0], lattice[1][0], lattice[2][0]),
    )
    fileobj.write(
        "%20.14f %20.14f %20.14f\n" % (lattice[3][0], lattice[4][0], lattice[5][0]),
    )
    fileobj.write(
        "%20.14f %20.14f %20.14f\n" % (lattice[6][0], lattice[7][0], lattice[8][0]),
    )

    elements = [s.species_string for s in structure.sites]
    elements_set = []
    elements_number = {}
    for e in elements:
        if e in elements_set:
            elements_number[e] = elements_number[e] + 1
        else:
            elements_set.append(e)
            elements_number[e] = 1

    for e in elements_set:
        fileobj.write("  " + e)
    fileobj.write("\n")

    for e in elements_set:
        fileobj.write("%5d " % (elements_number[e]))
    fileobj.write("\n")
    if idirect:
        fileobj.write("Direct\n")
        for p in structure.frac_coords:
            fileobj.write("%20.14f %20.14f %20.14f\n" % (p[0], p[1], p[2]))
    else:
        fileobj.write("Cartesian\n")
        for p in structure.cart_coords:
            fileobj.write("%20.14f %20.14f %20.14f\n" % (p[0], p[1], p[2]))


def _write_specific_format(
    structure,
    grid: Union["np.ndarray", Sequence[int]],
    volumetricData: Union["np.ndarray", Sequence[float]],
    filename: str,
    gridsize: Optional[Union["np.ndarray", Sequence[int]]] = None,
    format: Optional[str] = "cube",
    compact: bool = False,
    inorm: bool = False,
):
    absfile = handle_duplicated_output(os.path.abspath(filename))
    format_from_filename = filename.split(".")[-1] if "." in filename else None
    os.makedirs(os.path.dirname(absfile), exist_ok=True)

    import numpy as np

    if inorm is True:
        volumetricData /= np.sum(volumetricData)

    ngx, ngy, ngz = grid
    if gridsize is None:
        gridsize = grid
        interp_data = np.asarray(volumetricData)
    else:
        oldngx = np.linspace(0, 1, ngx)
        oldngy = np.linspace(0, 1, ngy)
        oldngz = np.linspace(0, 1, ngz)
        from scipy.interpolate import RegularGridInterpolator

        rgi = RegularGridInterpolator((oldngx, oldngy, oldngz), volumetricData)

        print("Interpolating volumetric data...")
        newngx = np.linspace(0, 1, gridsize[0])
        newngy = np.linspace(0, 1, gridsize[1])
        newngz = np.linspace(0, 1, gridsize[2])
        X, Y, Z = np.meshgrid(newngx, newngy, newngz, indexing="ij")
        points = np.asarray([X.ravel(), Y.ravel(), Z.ravel()]).T
        interp_data = rgi(points).reshape(gridsize)
        print("volumetric data interpolated")

    if format_from_filename and format_from_filename.lower() in [
        "vesta",
        "cube",
        "vasp",
        "txt",
    ]:
        _format = format_from_filename.lower()
    else:
        _format = format.lower() if format else "cube"

    if _format == "cube":
        volume_in_Bohr3 = structure.volume / Bohr**3
        interp_data /= volume_in_Bohr3
        with open(absfile, "w") as fileobj:
            import time

            fileobj.write("Cube file written on " + time.strftime("%c"))
            fileobj.write("\nOUTER LOOP: X, MIDDLE LOOP: Y, INNER LOOP: Z\n")

            origin = np.zeros(3)
            fileobj.write(
                f"{len(structure.sites):5d} {origin[0]:12.6f} {origin[1]:12.6f} {origin[2]:12.6f}\n",
            )

            for i in range(3):
                n = gridsize[i]
                d = structure.lattice.matrix[i] / n / Bohr
                fileobj.write(f"{n:5d} {d[0]:12.6f} {d[1]:12.6f} {d[2]:12.6f}\n")

            positions = structure.cart_coords / Bohr
            species_string = [s.species_string for s in structure.sites]
            species_string = [
                s.replace("+", "").replace("-", "") for s in species_string
            ]
            symbols = "".join(species_string)  # SiH
            from dspawpy.io.utils import _symbols2numbers

            numbers = _symbols2numbers(symbols)
            for Z, (x_index, y_index, z_index) in zip(numbers, positions):
                fileobj.write(
                    f"{Z:5d} {0:12.6f} {x_index:12.6f} {y_index:12.6f} {z_index:12.6f}\n",
                )

            if compact:
                for x_index in range(gridsize[0]):
                    for y_index in range(gridsize[1]):
                        for z_index in range(gridsize[2]):
                            fileobj.write(
                                f"{interp_data[x_index, y_index, z_index]:12.5e}\n",
                            )
            else:
                for x_index in range(gridsize[0]):
                    for y_index in range(gridsize[1]):
                        for z_index in range(gridsize[2]):
                            fileobj.write(
                                f"{interp_data[x_index, y_index, z_index]:12.5e} ",
                            )
                            if z_index % 6 == 5:
                                fileobj.write("\n")
                        fileobj.write("\n")

    elif _format == "vesta" or _format == "vasp":
        with open(absfile, "w") as file:
            _write_atoms(file, structure, idirect=True)
            file.write("%5d %5d %5d\n" % (gridsize[0], gridsize[1], gridsize[2]))
            count = 0
            if compact:
                interp_data.flatten().tofile(file, sep="\n", format="%.5e")
            else:
                for z_index in range(gridsize[2]):
                    for y_index in range(gridsize[1]):
                        for x_index in range(gridsize[0]):
                            file.write(
                                f"{interp_data[x_index, y_index, z_index]:12.5e} ",
                            )
                            count += 1
                            if count % 5 == 0:
                                file.write("\n")

    elif _format == "txt":
        with open(absfile, "w") as file:
            file.write(f"# {len(structure.sites)} atoms\n")
            file.write(f"# {gridsize[0]} {gridsize[1]} {gridsize[2]} grid size\n")
            file.write("# x y z value\n")
            for x_index in range(gridsize[0]):
                for y_index in range(gridsize[1]):
                    for z_index in range(gridsize[2]):
                        matrix = structure.lattice.matrix
                        v_x = matrix[0] * x_index / gridsize[0]
                        v_y = matrix[1] * y_index / gridsize[1]
                        v_z = matrix[2] * z_index / gridsize[2]
                        v_xyz = v_x + v_y + v_z
                        x, y, z = v_xyz
                        file.write(
                            f"{x:.3f} {y:.3f} {z:.3f} {interp_data[x_index, y_index, z_index]:14.7f}\n",
                        )
            import time

            file.write(f"# {format} file written on {time.strftime('%c')}\n")

    else:
        raise NotImplementedError('only "cube", "vesta", "vasp", "txt" are supported.')

    print(f"==> {absfile}")


def _to_dspaw_as(structure, filename: str, coords_are_cartesian: bool = True):
    """Write dspaw structure file of .as type"""

    absfile = handle_duplicated_output(os.path.abspath(filename))
    os.makedirs(os.path.dirname(absfile), exist_ok=True)
    with open(absfile, "w", encoding="utf-8") as file:
        file.write("Total number of atoms\n")
        file.write("%d\n" % len(structure))

        # ^ write lattice info
        if "LatticeFixs" in structure.sites[0].properties:
            lfinfo = structure.sites[0].properties["LatticeFixs"]
            if len(lfinfo) == 3:
                file.write("Lattice Fix\n")
                formatted_fts = []
                for ft in lfinfo:
                    if ft == "True":  # True
                        ft_formatted = "T"
                    else:
                        ft_formatted = "F"
                    formatted_fts.append(ft_formatted)
                for v in structure.lattice.matrix:
                    # write each element of formatted_fts in a line without [] symbol
                    file.write(f'{v} {formatted_fts}.strip("[").strip("]")\n')
            elif len(lfinfo) == 9:
                file.write("Lattice Fix_x Fix_y Fix_z\n")
                formatted_fts = []
                for ft in lfinfo:
                    if ft == "True":  # True
                        ft_formatted = "T"
                    else:
                        ft_formatted = "F"
                    formatted_fts.append(ft_formatted)
                fix_str1 = " ".join(formatted_fts[:3])
                fix_str2 = " ".join(formatted_fts[3:6])
                fix_str3 = " ".join(formatted_fts[6:9])
                v1 = structure.lattice.matrix[0]
                v2 = structure.lattice.matrix[1]
                v3 = structure.lattice.matrix[2]
                file.write(f" {v1[0]:5.8f} {v1[1]:5.8f} {v1[2]:5.8f} {fix_str1}\n")
                file.write(f" {v2[0]:5.8f} {v2[1]:5.8f} {v2[2]:5.8f} {fix_str2}\n")
                file.write(f" {v3[0]:5.8f} {v3[1]:5.8f} {v3[2]:5.8f} {fix_str3}\n")
            else:
                raise ValueError(
                    f"LatticeFixs should be a Sequence of 3 or 9 bools, but got {lfinfo}",
                )
        else:
            file.write("Lattice\n")
            for v in structure.lattice.matrix:
                file.write("%.8f %.8f %.8f\n" % (v[0], v[1], v[2]))

        i = 0
        for site in structure:
            keys = []
            for key in site.properties:  # site.properties is a dictionary
                if key != "LatticeFixs":
                    keys.append(key)
            keys.sort()
            keys_str = " ".join(keys)  # sth like 'magmom fix
            if i == 0:
                if coords_are_cartesian:
                    file.write(f"Cartesian {keys_str}\n")
                else:
                    file.write(f"Direct {keys_str}\n")
            i += 1

            coords = site.coords if coords_are_cartesian else site.frac_coords
            raw = []
            for sortted_key in keys:  # site.properties is a dictionary
                raw_values = site.properties[sortted_key]
                if not isinstance(raw_values, str) and isinstance(raw_values, Sequence):
                    values = raw_values
                else:
                    values = [raw_values]
                for v in values:
                    if v == "True":
                        value_str = "T"
                    elif v == "False":
                        value_str = "F"
                    else:
                        value_str = str(v)
                    raw.append(value_str)

            final_strs = " ".join(raw)  # sth like '0.0 T
            # remove all digits and +/- symbols
            sss = ""
            for char in site.species_string:
                if not char.isdigit() and char not in ["+", "-"]:
                    sss += char
            file.write(
                "%s %.8f %.8f %.8f %s\n"
                % (
                    sss,
                    coords[0],
                    coords[1],
                    coords[2],
                    final_strs,
                ),
            )
    print(f"==> {absfile}")


def _to_hzw(structure, filename: str):
    """Write hzw structure file of .hzw type"""

    absfile = handle_duplicated_output(os.path.abspath(filename))
    os.makedirs(os.path.dirname(absfile), exist_ok=True)
    with open(absfile, "w", encoding="utf-8") as file:
        file.write("% The number of probes \n")
        file.write("0\n")
        file.write("% Uni-cell vector\n")

        for v in structure.lattice.matrix:
            file.write("%.6f %.6f %.6f\n" % (v[0], v[1], v[2]))

        file.write("% Total number of device_structure\n")
        file.write("%d\n" % len(structure))
        file.write("% Atom site\n")

        for site in structure:
            file.write(
                "%s %.6f %.6f %.6f\n"
                % (site.species_string, site.coords[0], site.coords[1], site.coords[2]),
            )
    print(f"==> {absfile}")


def _to_dspaw_json(structure, filename: str, coords_are_cartesian: bool = True):
    """Write dspaw structure file of .json type"""

    absfile = handle_duplicated_output(os.path.abspath(filename))
    lattice = structure.lattice.matrix.flatten().tolist()
    atoms = []
    for site in structure:
        coords = site.coords if coords_are_cartesian else site.frac_coords
        atoms.append({"Element": site.species_string, "Position": coords.tolist()})

    coordinate_type = "Cartesian" if coords_are_cartesian else "Direct"
    d = {"Lattice": lattice, "CoordinateType": coordinate_type, "Atoms": atoms}
    os.makedirs(os.path.dirname(absfile), exist_ok=True)
    with open(absfile, "w", encoding="utf-8") as file:
        from json import dump

        dump(d, file, indent=4)
    print(f"==> {absfile}")


def _to_pdb(structures, filename: str):
    """Write pdb structure file of .pdb type"""

    absfile = handle_duplicated_output(os.path.abspath(filename))
    if not isinstance(structures, Sequence):
        structures = [structures]
    os.makedirs(os.path.dirname(absfile), exist_ok=True)
    with open(absfile, "w", encoding="utf-8") as file:
        for i, s in enumerate(structures):
            file.write("MODEL         %d\n" % (i + 1))
            file.write("REMARK   Converted from Structures\n")
            file.write("REMARK   Converted using dspawpy\n")
            # may lack lattice info
            if hasattr(s, "lattice"):
                lengths = s.lattice.lengths
                angles = s.lattice.angles
                file.write(
                    f"CRYST1{lengths[0]:9.3f}{lengths[1]:9.3f}{lengths[2]:9.3f}{angles[0]:7.2f}{angles[1]:7.2f}{angles[2]:7.2f}\n",
                )
            for j, site in enumerate(s):
                file.write(
                    "%4s%7d%4s%5s%6d%4s%8.3f%8.3f%8.3f%6.2f%6.2f%12s\n"
                    % (
                        "ATOM",
                        j + 1,
                        site.species_string,
                        "MOL",
                        1,
                        "    ",
                        site.coords[0],
                        site.coords[1],
                        site.coords[2],
                        1.0,
                        0.0,
                        site.species_string,
                    ),
                )
            file.write("TER\n")
            file.write("ENDMDL\n")

    print(f"==> {absfile}")


def _to_image(structures, filename: str, fmt: str = "png"):
    """将结构导出为图像格式（png/gif）

    Parameters
    ----------
    structures:
        pymatgen Structure对象或Structure列表
    filename:
        输出文件名
    fmt:
        图像格式，支持 'png' 或 'gif'
    """
    from pymatgen.io.ase import AseAtomsAdaptor

    absfile = handle_duplicated_output(os.path.abspath(filename))
    os.makedirs(os.path.dirname(absfile), exist_ok=True)

    # 转换为列表
    if not isinstance(structures, Sequence):
        structures = [structures]

    # 转换为ASE Atoms对象
    adaptor = AseAtomsAdaptor()
    ase_structures = [adaptor.get_atoms(s) for s in structures]

    # 使用ASE的io模块写入图像
    from ase.io import write as ase_write

    if fmt == "gif" and len(ase_structures) > 1:
        # 对于多个结构，创建gif动画
        ase_write(absfile, ase_structures, format="gif", rotation="10x,10y,10z")
    elif fmt == "png" or (fmt == "gif" and len(ase_structures) == 1):
        # 对于单个结构或png格式，只保存最后一个结构
        ase_write(absfile, ase_structures[-1], format="png", rotation="10x,10y,10z")
    else:
        raise ValueError(f"不支持的图像格式: {fmt}")

    print(f"==> {absfile}")



import os
import threading
from typing import Literal, Optional, Tuple

from loguru import logger
from prompt_toolkit import prompt
from prompt_toolkit.completion import (
    Completer,
    FuzzyCompleter,
    WordCompleter,
)

from dspawpy import auto_test_cli
from dspawpy.cli import list_ds, pc
from dspawpy.cli.menu_prompts import Dcheck, Dio, Dparameter, Dresponse, Dselect


def save_figure_dpi300(outfile: str, plt=None):
    """Save matplotlib figure with dpi=300"""
    logger.info(" Saving figure with dpi=300...")
    if plt is None:
        import matplotlib.pyplot as plt

    absfile = os.path.abspath(outfile)
    os.makedirs(os.path.dirname(absfile), exist_ok=True)
    plt.tight_layout()
    plt.savefig(absfile, dpi=300)
    logger.info(f"--> {absfile}")


def get_input(
    user_prompt: str,
    valid_inputs: list,
    completer: Optional[Completer] = None,
    allow_empty: bool = False,
    default_user_input: str = "",
) -> str:
    """Until user give valid input, or return default input if allow_empty is True and user input is empty."""
    if completer is None:
        completer = FuzzyCompleter(WordCompleter(valid_inputs))  # list completion
    from prompt_toolkit import prompt

    while True:
        user_input: str = prompt(user_prompt, completer=completer).strip()
        if allow_empty and len(user_input) == 0:
            return default_user_input
        elif user_input not in valid_inputs:
            continue
        else:
            return user_input


def get_inputs(
    user_prompt: str,
    valid_inputs: list,
    completer: Optional[Completer] = None,
    allow_empty: bool = False,
) -> list:
    """Return valid_inputs list if given empty"""
    if completer is None:
        completer = FuzzyCompleter(WordCompleter(valid_inputs))  # list completion
    from prompt_toolkit import prompt

    while True:
        user_inputs: list = prompt(user_prompt, completer=completer).strip().split()

        if allow_empty and len(user_inputs) == 0:
            return valid_inputs
        else:
            valid = True
            for n in user_inputs:
                if n not in valid_inputs:
                    valid = False
                    break
            if valid:
                return user_inputs
            else:
                continue


def get_lims(
    user_prompt: str,
    language: Literal["EN", "CN"],
) -> Optional[Tuple[float, float]]:
    """Get lower and higher limits from user input"""
    while True:
        userInput = input(user_prompt).strip()
        if userInput == "":
            return None
        elif " " not in userInput:
            logger.warning(f"!!! {userInput}{Dresponse[language][0]}")
            continue
        elif len(userInput.split(" ")) != 2:
            logger.warning(f"!!! {userInput}{Dresponse[language][1]}")
            continue
        else:
            try:
                lims = userInput.split(" ")
                lower = float(lims[0])
                higher = float(lims[1])
                return lower, higher

            except Exception:
                logger.warning(f"!!! {userInput}{Dresponse[language][2]}")
                continue


def online_check(local_dspawpy_version: str, language: Literal["EN", "CN"]):
    """Fetch latses version of dspawpy from pypi"""
    latest_version = None
    # requests is dependency of pymatgen
    from os.path import expanduser

    from requests import exceptions, get

    try:
        logger.info(Dcheck[language][0])
        response = get("https://pypi.org/pypi/dspawpy/json", timeout=3)
        latest_version = response.json()["info"]["version"]
        error_message = None
    except ModuleNotFoundError:
        error_message = Dcheck[language][1]
    except exceptions.Timeout:
        error_message = Dcheck[language][2]
    except exceptions.RequestException as e:
        error_message = f"{Dcheck[language][3]} {e}"
    except Exception as e:
        error_message = f"{Dcheck[language][4]} {e}"
    finally:
        if latest_version:
            if latest_version != local_dspawpy_version:
                logger.info(
                    f"{latest_version} > {local_dspawpy_version}; {Dcheck[language][1]}",
                )
            else:
                logger.info(
                    f"{latest_version} = {local_dspawpy_version}; {Dcheck[language][6]}",
                )

            with open(expanduser("~/.dspawpy_latest_version"), "w") as fin:
                fin.write(latest_version)
        else:
            logger.info(Dcheck[language][7])

    return error_message


def verify_dspawpy_version(check: bool, language: Literal["EN", "CN"]):
    """May skip online check"""
    from os.path import dirname, expanduser, isfile

    import dspawpy

    home = expanduser("~")

    dv = dspawpy.__version__
    df = dirname(dspawpy.__file__).replace(home, "~", 1)

    logger.info(f"{dv}: {df}")

    error_message = None
    if check:
        if isfile(expanduser("~/.dspawpy_latest_version")):
            with open(expanduser("~/.dspawpy_latest_version")) as fin:
                latest_version = fin.read().strip()
            if dv != latest_version:
                error_message = online_check(dv, language)
        else:
            error_message = online_check(dv, language)

        if error_message is not None:
            logger.info(error_message)


def s2(language: Literal["EN", "CN"]) -> dict:
    """Structure transformation"""

    if auto_test_cli:
        from dspawpy.io.structure import convert as cli_convert

        for D in list_ds:
            if D["menu"] == 2:
                cli_convert(infile=D["in"], outfile=D["out"])
        D = {}
    else:

        def imp():
            global convert
            from dspawpy.io.structure import convert

        import_thread = threading.Thread(target=imp)
        import_thread.start()

        D = {}
        D["menu"] = 2
        D["in"] = prompt(Dio[language]["ins"], completer=pc)
        D["out"] = prompt(Dio[language]["outs"], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])

        convert(infile=D["in"], outfile=D["out"])

    return D


def s7(language: Literal["EN", "CN"]) -> dict:
    """Data processing for optical properties"""
    if auto_test_cli:
        from os import makedirs as cli_makedirs

        from dspawpy.plot import plot_optical as cli_plot_optical

        for D in list_ds:
            if D["menu"] == 7:
                if D["outd"].strip() != "":
                    cli_makedirs(D["outd"], exist_ok=True)
                cli_plot_optical(
                    datafile=D["inf"], keys=D["keys"], axes=D["label"], prefix=D["outd"]
                )

        D = {}
    else:

        def imp():
            global plot_optical, makedirs
            from os import makedirs

            from dspawpy.plot import plot_optical

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 7
        D["inf"] = prompt(Dio[language]["optical"], completer=pc)
        _list = [
            "AbsorptionCoefficient",
            "ExtinctionCoefficient",
            "RefractiveIndex",
            "Reflectance",
        ]
        D["keys"] = get_inputs(
            Dselect[language][15],
            _list,
            completer=WordCompleter(_list),
            allow_empty=True,
        )
        _list2 = ["X", "Y", "Z", "XY", "YZ", "ZX"]
        D["label"] = get_inputs(
            Dselect[language][16],
            _list2,
            completer=WordCompleter(_list2),
            allow_empty=True,
        )

        D["outd"] = prompt(Dio[language]["outd"], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])

        if D["outd"].strip() != "":
            makedirs(D["outd"], exist_ok=True)
        plot_optical(
            datafile=D["inf"], keys=D["keys"], axes=D["label"], prefix=D["outd"]
        )

    return D


def s11(language: Literal["EN", "CN"]) -> dict:
    """Polarization iron electrode polarization data processing"""
    if auto_test_cli:
        from dspawpy.plot import (
            plot_polarization_figure as cli_plot_polarization_figure,
        )

        for D in list_ds:
            if D["menu"] == 7:
                cli_plot_polarization_figure(
                    D["inf"], D["rep"], figname=D["figure"], show=False
                )

        D = {}
    else:

        def imp():
            global plot_polarization_figure
            from dspawpy.plot import plot_polarization_figure

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 11
        D["inf"] = prompt(Dio[language]["polarization"], completer=pc)
        D["figure"] = prompt(Dio[language]["figure"], completer=pc)
        D["rep"] = int(input(Dparameter[language][5]) or 2)
        import_thread.join()
        logger.info(Dresponse[language][16])

        plot_polarization_figure(D["inf"], D["rep"], figname=D["figure"], show=False)

    return D


def s12(language: Literal["EN", "CN"]) -> dict:
    """Data processing for ZPE zero-point vibrational energy"""

    if auto_test_cli:
        from dspawpy.io.utils import getZPE as cli_getZPE

        for D in list_ds:
            if D["menu"] == 12:
                cli_getZPE(D["inf"])

        D = {}

    else:

        def imp():
            global getZPE
            from dspawpy.io.utils import getZPE

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 12
        D["inf"] = prompt(Dio[language]["txt"], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])

        getZPE(D["inf"])

    return D


def s14(language: Literal["EN", "CN"]) -> dict:
    if auto_test_cli:
        from dspawpy.analysis.post_relax import read_relax_data as cli_read_relax_data

        for D in list_ds:
            if D["menu"] == 14:
                cli_read_relax_data(D["logfile"])

        D = {}

    else:

        def imp():
            global read_relax_data
            from dspawpy.analysis.post_relax import read_relax_data

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 14
        D["logfile"] = prompt(Dio[language]["logfile"], completer=pc)
        if D["logfile"] == "":
            D["logfile"] = "DS-PAW.log"

        D["printef"] = True if prompt(Dselect[language][20]).startswith("y") else False
        D["relative"] = False
        if D["printef"]:
            D["relative"] = (
                True if prompt(Dselect[language][21]).startswith("y") else False
            )

        D["writetraj"] = (
            True if prompt(Dselect[language][23]).startswith("y") else False
        )
        D["xyzfilename"] = None
        D["writeas"] = True if prompt(Dselect[language][24]).startswith("y") else False

        import_thread.join()
        logger.info(Dresponse[language][16])

        read_relax_data(
            D["logfile"],
            D["printef"],
            D["relative"],
            D["writetraj"],
            D["writeas"],
        )
    return D



import threading
from typing import Literal

from loguru import logger
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from dspawpy import auto_test_cli
from dspawpy.cli import list_ds, pc
from dspawpy.cli.auxiliary import get_input, get_inputs, save_figure_dpi300
from dspawpy.cli.menu_prompts import Dio, Dresponse, Dselect


def s3_1(language: Literal["EN", "CN"]) -> dict:
    "Visualize volumetricData"
    if auto_test_cli:
        from dspawpy.io.write import write_VESTA as cli_write_VESTA

        for D in list_ds:
            if D["menu"] == 31:
                cli_write_VESTA(
                    in_filename=D["inf"],
                    data_type=D["task"],
                    out_filename=D["outfile"],
                    subtype=D["subtype"],
                )
        D = {}
    else:

        def imp():
            global write_VESTA
            from dspawpy.io.write import write_VESTA

        import_thread = threading.Thread(target=imp)
        import_thread.start()

        D = {}
        D["menu"] = 31
        D["inf"] = prompt(Dio[language]["inf"], completer=pc)
        supported_tasks = ["rho", "potential", "elf", "pcharge", "rhoBound"]
        D["task"] = None
        for task in supported_tasks:
            if task in D["inf"]:
                D["task"] = task
                break
        if not D["task"]:
            D["task"] = get_input(
                f"{supported_tasks}: ",
                supported_tasks,
                completer=WordCompleter(supported_tasks),
            )

        D["subtype"] = None
        if D["task"] == "potential":
            if D["inf"].endswith(".h5"):
                from dspawpy.io.read import load_h5

                data = load_h5(D["inf"])
                keys = [
                    k.split("/")[-1] for k in data.keys() if k.startswith("/Potential")
                ]

                if len(keys) == 0:
                    raise ValueError(f"{Dresponse[language][3]}{D['infile']}")
                elif len(keys) == 1:
                    D["subtype"] = keys[0]
                else:
                    D["subtype"] = get_input(
                        f"{Dselect[language][0]} {keys}: ",
                        keys,
                        completer=WordCompleter(keys),
                    )

            elif D["inf"].endswith(".json"):
                from json import load

                with open(D["inf"]) as fin:
                    data = load(fin)
                    if "Potential" not in data.keys():
                        raise ValueError(f"{Dresponse[language][3]} {D['infile']}")
                    keys = [k for k in data["Potential"].keys()]

                if len(keys) == 1:
                    D["subtype"] = keys[0]
                else:
                    D["subtype"] = get_input(
                        f"{Dselect[language][0]} {keys}",
                        keys,
                        completer=WordCompleter(keys),
                    )
            else:
                raise ValueError(Dresponse[language][2])

        D["outfile"] = prompt(Dio[language]["outf"], completer=pc)

        import_thread.join()
        logger.info(Dresponse[language][16])
        write_VESTA(  # type: ignore
            in_filename=D["inf"],
            data_type=D["task"],
            out_filename=D["outfile"],
            subtype=D["subtype"],
        )
        logger.info(Dresponse[language][5])

    return D


def s3_2(language: Literal["EN", "CN"]) -> dict:
    """Differential volumetricData visualization"""

    if auto_test_cli:
        D = {}
        from dspawpy.io.write import write_delta_rho_vesta as cli_write_delta_rho_vesta

        for D in list_ds:
            if D["menu"] == 32:
                cli_write_delta_rho_vesta(
                    total=D["total"],
                    individuals=D["individuals"],
                    output=D["outfile"],
                    data_type=D["task"],
                )
    else:

        def imp():
            global write_delta_rho_vesta
            from dspawpy.io.write import write_delta_rho_vesta

        import_thread = threading.Thread(target=imp)
        import_thread.start()

        D = {}
        D["menu"] = 32
        D["total"] = prompt(Dio[language]["tcharge"], completer=pc)
        D["individuals"] = []
        while True:
            individual = prompt(
                Dio[language]["pcharge"],
                completer=pc,
            )
            if individual == "":
                break
            D["individuals"].append(individual)

        supported_tasks = ["rho", "potential", "elf", "pcharge", "rhoBound"]
        D["task"] = None
        for task in supported_tasks:
            if task in D["total"]:
                D["task"] = task
                break
        if not D["task"]:
            D["task"] = get_input(
                f"{supported_tasks}: ",
                supported_tasks,
                completer=WordCompleter(supported_tasks),
            )
        D["outfile"] = prompt(Dio[language]["outf"], completer=pc)

        import_thread.join()
        logger.info(Dresponse[language][16])
        write_delta_rho_vesta(
            total=D["total"],
            individuals=D["individuals"],
            output=D["outfile"],
            data_type=D["task"],
        )
        logger.info(Dresponse[language][5])

    return D


def s3_3(language: Literal["EN", "CN"]) -> dict:
    "Volumetric data mean"
    if auto_test_cli:
        import matplotlib.pyplot as cli_plt

        from dspawpy.plot import average_along_axis as cli_average_along_axis

        D = {}
        for D in list_ds:
            if D["menu"] == 33:
                cli_plt = cli_average_along_axis(D["inf"], D["axes"])
                cli_plt.ylim(D["ylims"])
                save_figure_dpi300(D["figure"], cli_plt)
    else:

        def imp():
            global plt, average_along_axis

            import matplotlib.pyplot as plt

            from dspawpy.plot import average_along_axis

        import_thread = threading.Thread(target=imp)
        import_thread.start()

        D = {}
        D["menu"] = 33
        D["inf"] = prompt(Dio[language]["inf"], completer=pc)
        D["axes"] = get_inputs(Dselect[language][1], ["0", "1", "2"])
        _list = ["rho", "potential", "elf", "pcharge", "rhoBound"]
        D["task"] = get_input(
            f"{_list}: ",
            _list,
            completer=WordCompleter(_list),
        )

        D["subtype"] = None
        if D["task"] == "rho":
            k = "TotalCharge"
        elif D["task"] == "potential":
            if D["inf"].endswith(".h5"):
                from dspawpy.io.read import load_h5

                data = load_h5(D["inf"])
                keys = [
                    k.split("/")[-1] for k in data.keys() if k.startswith("/Potential")
                ]

                if len(keys) == 0:
                    raise ValueError(f"{Dresponse[language][3]}{D['infile']}")
                elif len(keys) == 1:
                    D["subtype"] = keys[0]
                else:
                    D["subtype"] = get_input(
                        Dselect[language][0],
                        keys,
                        completer=WordCompleter(keys),
                    )

            elif D["inf"].endswith(".json"):
                from json import load

                with open(D["inf"]) as fin:
                    data = load(fin)
                    if "Potential" not in data.keys():
                        raise ValueError(f"{Dresponse[language][3]}{D['infile']}")
                    keys = [k for k in data["Potential"].keys()]

                if len(keys) == 1:
                    D["subtype"] = keys[0]
                else:
                    D["subtype"] = get_input(
                        Dselect[language][0],
                        keys,
                        completer=WordCompleter(keys),
                    )
            else:
                raise ValueError(Dresponse[language][4])
            k = D["subtype"]
        elif D["task"] == "elf":
            k = "TotalELF"
        elif D["task"] == "pcharge":
            k = "TotalCharge"
        elif D["task"] == "rhoBound":
            k = "Rho"
        else:
            raise ValueError(D["task"])

        ax_indices = [int(a) for a in D["axes"]]
        import_thread.join()
        logger.info(Dresponse[language][16])
        for ai in ax_indices:
            average_along_axis(
                datafile=D["inf"],
                task=D["task"],
                axis=ai,
                subtype=D["subtype"],
                label=f"axis{ai}",
            )
        if len(ax_indices) > 1:
            plt.legend()

        plt.xlabel("Grid Index")
        plt.ylabel(k)
        D["figure"] = prompt(Dio[language]["figure"], completer=pc)

        save_figure_dpi300(D["figure"], plt)

    return D



import os
import sys
import threading
from typing import Literal

from loguru import logger
from prompt_toolkit import prompt

from dspawpy import auto_test_cli
from dspawpy.cli import list_ds, pc
from dspawpy.cli.auxiliary import get_lims, save_figure_dpi300
from dspawpy.cli.menu_prompts import Dio, Dparameter, Dresponse


class HidePrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def pre_ph_band(language: Literal["EN", "CN"]):
    if auto_test_cli:
        from dspawpy.io.read import (
            get_phonon_band_data as cli_get_phonon_band_data,
        )

        with HidePrints():
            from pymatgen.phonon.plotter import (
                PhononBSPlotter as cli_PhononBSPlotter,
            )

        for D in list_ds:
            if D["menu"] == 91:
                cli_band_data = cli_get_phonon_band_data(D["inf"])
                cli_bsp = cli_PhononBSPlotter(cli_band_data)
                cli_bsp.get_plot(ylim=D["ylims"])  # pyright: ignore [reportArgumentType]
                save_figure_dpi300(D["figure"])

    def imp():
        global get_phonon_band_data, PhononBSPlotter
        from dspawpy.io.read import get_phonon_band_data

        with HidePrints():
            from pymatgen.phonon.plotter import PhononBSPlotter

    import_thread = threading.Thread(target=imp)
    import_thread.start()

    D = {}
    D["inf"] = prompt(Dio[language]["phband"], completer=pc)
    D["ylims"] = get_lims(Dparameter[language][10], language)
    D["figure"] = prompt(Dio[language]["figure"], completer=pc) or "ph_band.png"

    import_thread.join()
    logger.info(Dresponse[language][16])

    return D, get_phonon_band_data, PhononBSPlotter


def s9_1(language: Literal["EN", "CN"]) -> dict:
    """Phonon band data processing"""

    if auto_test_cli:
        from pymatgen.phonon.plotter import PhononBSPlotter as cli_PhononBSPlotter

        from dspawpy.io.read import get_phonon_band_data as cli_get_phonon_band_data

        for D in list_ds:
            if D["menu"] == 91:
                band_data = cli_get_phonon_band_data(D["inf"])
                bsp = cli_PhononBSPlotter(band_data)
                bsp.get_plot(ylim=D["ylims"])  # pyright: ignore [reportArgumentType]
                save_figure_dpi300(D["figure"])
        D = {}

    else:
        D, get_phonon_band_data, PhononBSPlotter = pre_ph_band(language)
        D["menu"] = 91

        band_data = get_phonon_band_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        bsp = PhononBSPlotter(band_data)
        logger.debug("plotter initialized")
        bsp.get_plot(ylim=D["ylims"])  # pyright: ignore [reportArgumentType]
        save_figure_dpi300(D["figure"])

    return D


def pre_ph_dos(language: Literal["EN", "CN"]):
    def imp():
        global PhononDosPlotter, get_phonon_dos_data
        from dspawpy.io.read import get_phonon_dos_data

        with HidePrints():
            from pymatgen.phonon.plotter import PhononDosPlotter

    import_thread = threading.Thread(target=imp)
    import_thread.start()

    D = {}
    D["inf"] = prompt(Dio[language]["phdos"], completer=pc)
    D["xlims"] = get_lims(Dparameter[language][9], language)
    D["ylims"] = get_lims(Dparameter[language][10], language)
    D["figure"] = prompt(Dio[language]["figure"], completer=pc) or "ph_dos.png"
    import_thread.join()
    logger.info(Dresponse[language][16])

    return D, get_phonon_dos_data, PhononDosPlotter


def s9_2(language: Literal["EN", "CN"]) -> dict:
    """Phonon density of states data processing"""
    if auto_test_cli:
        from pymatgen.phonon.plotter import PhononDosPlotter as cli_PhononDosPlotter

        from dspawpy.io.read import get_phonon_dos_data as cli_get_phonon_dos_data

        for D in list_ds:
            if D["menu"] == 92:
                dos_data = cli_get_phonon_dos_data(D["inf"])
                dos_plotter = cli_PhononDosPlotter(stack=False, sigma=None)
                dos_plotter.add_dos(label="Phonon", dos=dos_data)
                dos_plotter.get_plot(
                    xlim=D["xlims"],  # pyright: ignore [reportArgumentType]
                    ylim=D["ylims"],  # pyright: ignore [reportArgumentType]
                    units="thz",
                )
                save_figure_dpi300(D["figure"])
        D = {}

    else:
        D, get_phonon_dos_data, PhononDosPlotter = pre_ph_dos(language)
        D["menu"] = 92

        dos_data = get_phonon_dos_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        dos_plotter = PhononDosPlotter(stack=False, sigma=None)
        logger.debug("plotter initialized")

        dos_plotter.add_dos(label="Phonon", dos=dos_data)
        dos_plotter.get_plot(
            xlim=D["xlims"],  # pyright: ignore [reportArgumentType]
            ylim=D["ylims"],  # pyright: ignore [reportArgumentType]
            units="thz",
        )

        save_figure_dpi300(D["figure"])

    return D


def s9_3(language: Literal["EN", "CN"]) -> dict:
    "Phonon thermodynamic data processing"
    if auto_test_cli:
        from dspawpy.plot import plot_phonon_thermal as cli_plot_phonon_thermal

        for D in list_ds:
            if D["menu"] == 93:
                cli_plot_phonon_thermal(D["inf"], D["figure"], True)
        D = {}

    else:

        def imp():
            global plot_phonon_thermal
            from dspawpy.plot import plot_phonon_thermal

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 93
        D["inf"] = prompt(Dio[language]["inf"], completer=pc)
        D["figure"] = prompt(Dio[language]["figure"], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])

        plot_phonon_thermal(D["inf"], D["figure"], False)

    return D



import os
import sys

import polars as pl
from loguru import logger

pl.Config.set_tbl_hide_column_data_types(True)

# allow expand all rows and columns
if os.environ.get("ALWAYS_EXPAND") == "True":
    pl.Config.set_tbl_rows(-1)
    pl.Config.set_tbl_cols(-1)


if os.environ.get("dspawpy_cli_test") == "manual":
    append_json = True
    auto_test_cli = False
elif os.environ.get("dspawpy_cli_test") == "auto":
    append_json = False
    auto_test_cli = True
else:
    append_json = False
    auto_test_cli = False


__version__ = "1.8.2"
assert sys.version_info >= (3, 9)

logger.remove()
level = os.getenv("DLL")  # dspawpy log level
if level:  # default to simulate no log
    logger.add(
        sys.stderr,
        level=level,
        format="<green>{time:MM-DD HH:mm:ss:SSS}</green> | <level>{message}</level>",
    )
else:
    logger.add(
        sys.stderr,
        colorize=True,
        format="<level>{message}</level>",
    )

logger.add(
    ".dspawpy.log",
    level="DEBUG",
    rotation="1 day",
    retention="1 week",
    compression="zip",
)



from typing import TYPE_CHECKING, List, Optional, Union

from loguru import logger

from dspawpy.io.utils import reader

if TYPE_CHECKING:
    from pymatgen.core.structure import Structure


def build_Structures_from_datafile(
    datafile: Union[str, List[str]],
    si=None,
    ele=None,
    ai=None,
    fmt=None,
    task="scf",
):
    """Deprecated alias to read"""
    logger.warning(
        "build_Structures_from_datafile is deprecated; use read instead",
        DeprecationWarning,
    )
    return read(datafile, si=si, ele=ele, ai=ai, fmt=fmt, task=task)


def _get_structure_list(
    df: str,
    si=None,
    ele=None,
    ai=None,
    fmt: Optional[str] = None,
    task: Optional[str] = "scf",
    verbose: bool = True,
):
    """Get pymatgen structures from single datafile.

    Parameters
    ----------
    df:
        Path to the data file or the folder containing the data file

    Returns
    -------
    List[Structure] : list of pymatgen structures

    """
    if task is None:
        task = "scf"

    import os

    if os.path.isdir(df) or df.endswith(".h5") or df.endswith(".json"):
        from dspawpy.io.utils import get_absfile

        absfile = get_absfile(df, task=task, verbose=verbose)
    else:  # for other type of datafile, such as .as, .hzw, POSCAR
        absfile = os.path.abspath(df)

    if fmt is None:
        fmt = absfile.split(".")[-1]
    else:
        assert isinstance(fmt, str)

    if fmt == "as":
        strs = [_from_dspaw_as(absfile)]
    elif fmt == "hzw":
        logger.warning("build from .hzw may lack mag & fix info!", category=UserWarning)
        strs = [_from_hzw(absfile)]
    elif fmt == "xyz":
        strs = [_from_xyz(absfile)]
    elif fmt == "pdb":
        strs = _from_pdb(absfile)
    elif fmt == "h5":
        from dspawpy.io.read import get_sinfo

        Nstep, elements, positions, lattices, D_mag_fix = get_sinfo(
            datafile=absfile,
            si=si,
            ele=ele,
            ai=ai,
        )  # returned positions, not scaled-positions
        # remove _ from elements
        import re

        elements = [re.sub(r"_", "", e) for e in elements]

        strs = []
        from pymatgen.core import Structure

        for i in range(Nstep):
            if D_mag_fix:
                strs.append(
                    Structure(
                        lattices[i],
                        elements,
                        positions[i],
                        coords_are_cartesian=True,
                        site_properties={k: v[i] for k, v in D_mag_fix.items()},
                    ),
                )
            else:
                strs.append(
                    Structure(
                        lattices[i],
                        elements,
                        positions[i],
                        coords_are_cartesian=True,
                    ),
                )

    elif fmt == "json":
        try:
            from dspawpy.io.read import get_sinfo

            Nstep, elements, positions, lattices, D_mag_fix = get_sinfo(
                datafile=absfile,
                si=si,
                ele=ele,
                ai=ai,
            )  # returned positions, not scaled-positions
            # remove _ from elements

            import re

            elements = [re.sub(r"_", "", e) for e in elements]

            strs = []
            from pymatgen.core import Structure

            for i in range(Nstep):
                if D_mag_fix:
                    strs.append(
                        Structure(
                            lattices[i],
                            elements,
                            positions[i],
                            coords_are_cartesian=True,
                            site_properties={k: v[i] for k, v in D_mag_fix.items()},
                        ),
                    )
                else:
                    strs.append(
                        Structure(
                            lattices[i],
                            elements,
                            positions[i],
                            coords_are_cartesian=True,
                        ),
                    )
        except Exception:  # try parse json with pymatgen
            from pymatgen.core import Structure

            strs = [Structure.from_file(absfile)]

    else:
        from pymatgen.core import Structure

        strs = [Structure.from_file(absfile)]

    return strs


def _from_dspaw_as(as_file: str = "structure.as") -> "Structure":
    """Read structure information from a DSPAW as structure file.

    Parameters
    ----------
    as_file:
        DSPAW's as structure file, default 'structure.as'

    Returns
    -------
    Structure
        a pymatgen Structure object

    """
    import os

    absfile = os.path.abspath(as_file)
    from dspawpy.io.read import get_lines_without_comment

    lines = get_lines_without_comment(absfile, "#")
    N = int(lines[1])  # number of atoms

    # parse lattice info
    lattice = []  # lattice matrix
    for line in lines[3:6]:
        vector = line.split()
        lattice.extend([float(vector[0]), float(vector[1]), float(vector[2])])
    import numpy as np

    lattice = np.asarray(lattice).reshape(3, 3)

    lat_fixs = []
    if lines[2].strip() != "Lattice":  # fix lattice
        lattice_fix_info = lines[2].strip().split()[1:]
        if lattice_fix_info == ["Fix_x", "Fix_y", "Fix_z"]:
            # ONLY support xyz fix in sequence, yzx will cause error
            for line in lines[3:6]:
                lfs = line.strip().split()[3:6]
                for lf in lfs:
                    if lf.startswith("T"):
                        lat_fixs.append("True")
                    elif lf.startswith("F"):
                        lat_fixs.append("False")
        elif lattice_fix_info == ["Fix"]:
            for line in lines[3:6]:
                lf = line.strip().split()[3]
                if lf.startswith("T"):
                    lat_fixs.append("True")
                elif lf.startswith("F"):
                    lat_fixs.append("False")
        else:
            raise ValueError("Lattice fix info error!")

    elements = []
    positions = []
    for i in range(N):
        atom = lines[i + 7].strip().split()
        elements.append(atom[0])
        positions.extend([float(atom[1]), float(atom[2]), float(atom[3])])

    mf_info = None
    l6 = lines[6].strip()  # str, 'Cartesian/Direct Mag Fix_x ...'
    if l6.split()[0] == "Direct":
        is_direct = True
    elif l6.split()[0] == "Cartesian":
        is_direct = False
    else:
        raise ValueError("Structure file format error!")

    mf_info = l6.split()[1:]  # ['Mag', 'Fix_x', 'Fix_y', 'Fix_z']
    for item in mf_info:
        assert item in [
            "Mag",
            "Mag_x",
            "Mag_y",
            "Mag_z",
            "Fix",
            "Fix_x",
            "Fix_y",
            "Fix_z",
        ], (
            f"{item} is not a valid flag! Expecting ['Mag', 'Mag_x', 'Mag_y', 'Mag_z', 'Fix', 'Fix_x', 'Fix_y', 'Fix_z']"
        )

    mag_fix_dict = {}
    if mf_info is not None:
        after_Fix = False
        for mf_index, item in enumerate(mf_info):
            values = []
            for i in range(N):
                atom = lines[i + 7].strip().split()
                mf = atom[4:]
                if item == "Fix":  # Fix == Fix_x, Fix_y, Fix_z
                    values.append([mf[mf_index], mf[mf_index + 1], mf[mf_index + 2]])
                elif item.startswith("Fix_"):
                    values.append(mf[mf_index])
                else:  # mag
                    if after_Fix:
                        values.append(float(mf[mf_index + 2]))
                    else:
                        values.append(float(mf[mf_index]))
            # set after_Fix flag, to shift index with 2
            if item == "Fix":
                after_Fix = True
            else:
                after_Fix = False

            if item.startswith("Fix"):  # F -> False, T -> True
                for value in values:
                    if isinstance(value, str):
                        if value.startswith("T"):
                            values[values.index(value)] = "True"
                        elif value.startswith("F"):
                            values[values.index(value)] = "False"
                    elif isinstance(value, list):
                        for v in value:
                            if v.startswith("T"):
                                value[value.index(v)] = "True"
                            elif v.startswith("F"):
                                value[value.index(v)] = "False"
            mag_fix_dict[item] = values
    if lat_fixs != []:
        # replicate lat_fixs to N atoms
        mag_fix_dict["LatticeFixs"] = [lat_fixs for _ in range(N)]

    coords = np.asarray(positions).reshape(-1, 3)
    # remove _ from elements
    import re

    elements = [re.sub(r"_", "", e) for e in elements]

    from pymatgen.core import Structure

    if mag_fix_dict == {}:
        return Structure(
            lattice,
            elements,
            coords,
            coords_are_cartesian=(not is_direct),
        )
    else:
        return Structure(
            lattice,
            elements,
            coords,
            coords_are_cartesian=(not is_direct),
            site_properties=mag_fix_dict,
        )


def _from_hzw(hzw_file) -> "Structure":
    """Read structure information from an hzw structure file

    Parameters
    ----------
    hzw_file:
        hzw structure file, with a .hzw extension

    Returns
    -------
    Structure
        pymatgen's Structure object

    Examples
    --------
    >>> from dspawpy.io.structure import _from_hzw
    >>> print(_from_hzw('tests/supplement/Si2.hzw'))
    Full Formula (Si2)
    Reduced Formula: Si
    abc   :   0.678839   0.678839   0.678839
    angles:  90.000000  90.000000  90.000000
    pbc   :       True       True       True
    Sites (2)
      #  SP           a         b         c
    ---  ----  --------  --------  --------
      0  Si    0.999999  0.999999  0.999999
      1  Si    3         3         3

    """
    import os

    from pymatgen.core import Structure

    from dspawpy.io.read import get_lines_without_comment

    absfile = os.path.abspath(hzw_file)

    lines = get_lines_without_comment(absfile, "%")
    number_of_probes = int(lines[0])
    elements = []
    positions = []
    if number_of_probes == 0:  # with lattice
        N = int(lines[4])
        for i in range(N):
            atom = lines[i + 5].strip().split()
            elements.append(atom[0])
            positions.append([float(atom[1]), float(atom[2]), float(atom[3])])

        lattice = []
        for line in lines[1:4]:
            vector = line.split()
            lattice.append([float(vector[0]), float(vector[1]), float(vector[2])])

    elif number_of_probes == -1:  # crystal
        N = int(lines[1])
        for i in range(N):
            atom = lines[i + 2].strip().split()
            elements.append(atom[0])
            positions.append([float(atom[1]), float(atom[2]), float(atom[3])])

        max_a = max(positions[:][0]) + 1e-6
        max_b = max(positions[:][0]) + 1e-6
        max_c = max(positions[:][0]) + 1e-6
        lattice = [max_a, 0, 0, 0, max_b, 0, 0, 0, max_c]

    else:
        raise NotImplementedError("number of probes must be 0 or 1")

    return Structure(lattice, elements, positions, coords_are_cartesian=True)


def _from_xyz(xyzfile: str) -> "Structure":
    """From molecule xyz file, build pmg.structure

    Examples
    --------
    >>> from dspawpy.io.structure import _from_xyz
    >>> print(_from_xyz('tests/supplement/Si2.xyz'))
    Full Formula (Si2)
    Reduced Formula: Si
    abc   :   4.811495   4.826272   4.826271
    angles:  90.000000  90.000000  90.000000
    pbc   :       True       True       True
    Sites (2)
      #  SP           a         b         c
    ---  ----  --------  --------  --------
      0  Si    0.928452  0.927353  0.927353
      1  Si    0.071548  0.072647  0.072647

    """
    from pymatgen.core.structure import Molecule

    mol = Molecule.from_file(xyzfile)
    assert mol is not None

    a = max(mol.cart_coords[:, 0]) + 1e-6
    b = max(mol.cart_coords[:, 1]) + 1e-6
    c = max(mol.cart_coords[:, 2]) + 1e-6

    return mol.get_boxed_structure(a, b, c, no_cross=True)


def _read_atom_line(line_full):
    """Read atomic name, xyz coordinates, and element symbol from a PDB file
    The format of the ATOM section in a PDB file is as follows (excluding magnetic moment and coordinate information):
    HETATM    1  H14 ORTE    0       6.301   0.693   1.919  1.00  0.00        H
    Fixed label Atom number Atom name Residue name Residue number x y z Occupancy Temperature factor Element symbol

    Exclude redundant information, retain only
        - Atom name
        - xyz coordinates
        - Element symbol
    """
    line = line_full.rstrip("\n")
    type_atm = line[0:6]
    if type_atm == "ATOM  " or type_atm == "HETATM":
        name = line[12:16].strip()  # Atom name
        # atomic coordinates
        import numpy as np

        try:  # 5.3f indicates 5 digits for the integer part and 3 digits for the decimal part, unit is Angstrom
            coord = np.asarray(
                [float(line[30:38]), float(line[38:46]), float(line[46:54])],
                dtype=np.float64,
            )
        except ValueError:
            raise ValueError("Invalid or missing coordinate(s)")
        symbol = line[76:78].strip().upper()  # Element symbol

    else:
        raise ValueError("Only ATOM and HETATM supported")

    return name, coord, symbol


@reader
def _from_pdb(fileobj):
    """Read PDB files. Modified from ASE"""
    images = []  # Conformation list
    import numpy as np

    orig = np.identity(3)  # Origin
    trans = np.zeros(3)  # Offset
    symbols = []  # Elements
    positions = []  # Coordinates
    cell = None  # Unit cell

    from pymatgen.core.lattice import Lattice

    for line in fileobj.readlines():
        if line.startswith("CRYST1"):
            cellpar = [
                float(line[6:15]),  # a
                float(line[15:24]),  # b
                float(line[24:33]),  # c
                float(line[33:40]),  # alpha
                float(line[40:47]),  # beta
                float(line[47:54]),
            ]  # gamma
            cell = Lattice.from_parameters(
                a=cellpar[0],
                b=cellpar[1],
                c=cellpar[2],
                alpha=cellpar[3],
                beta=cellpar[4],
                gamma=cellpar[5],
            )

        for c in range(3):
            if line.startswith("ORIGX" + "123"[c]):
                orig[c] = [float(line[10:20]), float(line[20:30]), float(line[30:40])]
                trans[c] = float(line[45:55])

        if line.startswith("ATOM") or line.startswith("HETATM"):
            # line_info = name, coord, symbol
            line_info = _read_atom_line(line)

            from dspawpy.io.utils import label_to_symbol

            try:  # Attempt to convert from element symbol, use atomic name if failure occurs
                symbol = label_to_symbol(line_info[2])
            except (KeyError, IndexError):
                symbol = label_to_symbol(line_info[0])
            symbols.append(symbol)

            position = np.dot(orig, line_info[1]) + trans
            positions.append(position)

        if line.startswith("END"):
            atoms = _build_atoms(cell, symbols, positions)
            images.append(atoms)
            symbols = []
            positions = []
            cell = None

    if len(images) == 0:
        atoms = _build_atoms(cell, symbols, positions)
        images.append(atoms)

    return images


def _build_atoms(cell, symbols, positions):
    if cell is None:
        logger.warning(
            "No lattice info in PDB file! The lattice defaults to [[2xmax, 0, 0]; [0, 2ymax, 0]; [0, 0, 2zmax]])",
            category=UserWarning,
        )
        # cell = np.zeros(shape=(3, 3))
        import numpy as np

        max_xyz = np.max(positions, axis=0)
        cell = np.diag(max_xyz * 2)

    from pymatgen.core import Structure

    atoms = Structure(
        lattice=cell,
        species=symbols,
        coords=positions,
        coords_are_cartesian=True,
    )

    return atoms


def read(
    datafile: Union[str, list],
    si=None,
    ele=None,
    ai=None,
    fmt: Optional[str] = None,
    task: Optional[str] = "scf",
):
    r"""Read one or more h5/json files and return a list of pymatgen Structures.

    Parameters
    ----------
    datafile:
        - file paths for h5/json/as/hzw/cif/poscar/cssr/xsf/mcsqs/prismatic/yaml/fleur-inpgen files;
        - If a directory path is given, it can be combined with the task parameter to read the {task}.h5/json files inside
        - If a list of strings is given, it will sequentially read the data and merge them into a list of Structures
    si: int, list or str
        - Configuration number, starting from 1

            - si=1, reads the first configuration
            - si=[1,2], reads the first and second configurations
            - si=':', reads all configurations
            - si='-3:', reads the last three configurations
        - If empty, it reads all configurations for multi-configuration files and the latest configuration for single-configuration files
        - This parameter is only valid for h5/json files
    ele:
        - Element symbol, format reference: 'H' or ['H','O']
        - If empty, it will read atomic information for all elements
        - This parameter is only valid for h5/json files
    ai:
        - Atom index, starting from 1
        - Same as si
        - If empty, it will read all atom information
        - This parameter is only valid for h5/json files
    fmt:
        - File format, including 'as', 'hzw', 'xyz', 'pdb', 'h5', 'json' 6 types, other values will be ignored.
        - If empty, the file type will be determined based on file name conventions.
    task:
        - Used when datafile is a directory path to find the internal {task}.h5/json file.
        - Determine the task type, including 'scf', 'relax', 'neb', 'aimd' four types, other values will be ignored.

    Returns
    -------
    pymatgen_Structures:
        Structure list

    Examples
    --------
    >>> from dspawpy.io.structure import read

    Reads a single file to generate a list of Structures

    >>> pymatgen_Structures = read(datafile='tests/supplement/PtH.as')
    >>> len(pymatgen_Structures)
    1
    >>> pymatgen_Structures = read(datafile='tests/supplement/PtH.hzw')
    >>> len(pymatgen_Structures)
    1
    >>> pymatgen_Structures = read(datafile='tests/supplement/Si2.xyz')
    >>> len(pymatgen_Structures)
    1
    >>> pymatgen_Structures = read(datafile='tests/aimd.pdb')
    >>> len(pymatgen_Structures)
    1000
    >>> pymatgen_Structures = read(datafile='tests/2.1/relax.h5')
    >>> len(pymatgen_Structures)
    3
    >>> pymatgen_Structures = read(datafile='tests/2.1/relax.json')
    >>> len(pymatgen_Structures)
    3

    Note that pymatgen_Structures is a list composed of multiple Structure objects, each corresponding to a structure. If there is only one structure, it will also return a list. Please use pymatgen_Structures[0] to obtain the Structure object.

    When datafile is a list, it reads multiple files sequentially and merges them into a Structures list

    >>> pymatgen_Structures = read(datafile=['tests/supplement/aimd1.h5','tests/supplement/aimd2.h5'])

    """
    dfs = []
    if isinstance(datafile, list):  # Continuation mode, multiple files are provided
        dfs = datafile
    else:  # Single calculation mode, processing a single file
        dfs.append(datafile)

    # Read structure data
    pymatgen_Structures = []
    for df in dfs:
        structure_list = _get_structure_list(df, si, ele, ai, fmt, task)
        pymatgen_Structures.extend(structure_list)

    return pymatgen_Structures


def write(
    structure,
    filename: str,
    fmt: Optional[str] = None,
    coords_are_cartesian: bool = True,
):
    r"""Write information to the structure file

    Parameters
    ----------
    structure:
        `A pymatgen Structure object`
    filename:
        Structure filename
    fmt:
        - Structure file type, natively supports 'json', 'as', 'hzw', 'pdb', 'xyz', 'dump', 'png', 'gif' eight types
    coords_are_cartesian:
        - Whether to write in Cartesian coordinates, default is True; otherwise write in fractional coordinate format
        - This option is currently only effective for 'as' and 'json' formats

    Examples
    --------
    First, read the structure information:

    >>> from dspawpy.io.structure import read, write
    >>> s = read('tests/2.15/01/neb01.h5')
    >>> len(s)
    17

    Writing structure information to a file:

    >>> write(s, filename='tests/doctest_out/PtH.json', coords_are_cartesian=True) # doctest: +ELLIPSIS
    ==> .../PtH...json
    >>> write(s, filename='tests/doctest_out/PtH.as', coords_are_cartesian=True) # doctest: +ELLIPSIS
    ==> .../PtH...as
    >>> write(s, filename='tests/doctest_out/PtH.hzw', coords_are_cartesian=True) # doctest: +ELLIPSIS
    ==> .../PtH...hzw

    PDB, XYZ, and DUMP file types can write multiple conformations to form a "trajectory." The generated XYZ trajectory files can be opened and visualized using visualization software like OVITO.

    >>> write(s, filename='tests/doctest_out/PtH.pdb', coords_are_cartesian=True) # doctest: +ELLIPSIS
    ==> .../PtH...pdb
    >>> write(s, filename='tests/doctest_out/PtH.xyz', coords_are_cartesian=True) # doctest: +ELLIPSIS
    ==> .../PtH...xyz
    >>> write(s, filename='tests/doctest_out/PtH.dump', coords_are_cartesian=True) # doctest: +ELLIPSIS
    ==> .../PtH...dump

    PNG and GIF formats can be used to export structure images using ASE's io module. PNG format exports a single structure image, while GIF format can create an animation for multiple structures.

    >>> write(s, filename='tests/doctest_out/PtH.png') # doctest: +ELLIPSIS
    ==> .../PtH...png
    >>> write(s, filename='tests/doctest_out/PtH.gif') # doctest: +ELLIPSIS
    ==> .../PtH...gif

    The recommended format for storing single structure information is as format. If the Structure contains magnetic moment or degree of freedom information, it will be written in the most complete format, such as Fix_x, Fix_y, Fix_z, Mag_x, Mag_y, Mag_z. The default value for degree of freedom information is F, and the default value for magnetic moment is 0.0. You can manually delete this default information from the generated as file as needed. Writing to other types of structure files will ignore magnetic moment and degree of freedom information.

    """
    from pymatgen.core import Structure

    if isinstance(structure, Structure):
        structure = [structure]

    import os

    absfilename = os.path.abspath(filename)
    if fmt is None:
        fmt = absfilename.split(".")[-1]

    if fmt == "pdb":  # Can be multiple conformations
        from .write import _to_pdb

        _to_pdb(structure, absfilename)
    elif fmt == "xyz":  # Can be multiple configurations
        from .write import _write_xyz_traj

        _write_xyz_traj(structure, absfilename)
    elif fmt == "dump":  # Can be multiple configurations
        from .write import _write_dump_traj

        _write_dump_traj(structure, absfilename)

    elif fmt == "json":  # Single configuration
        from .write import _to_dspaw_json

        _to_dspaw_json(structure[-1], absfilename, coords_are_cartesian)
    elif fmt == "as":
        from .write import _to_dspaw_as

        _to_dspaw_as(structure[-1], absfilename, coords_are_cartesian)
    elif fmt == "hzw":
        from .write import _to_hzw

        _to_hzw(structure[-1], absfilename)
    elif fmt == "png":
        from .write import _to_image

        _to_image(structure, absfilename, fmt="png")
    elif fmt == "gif":
        from .write import _to_image

        _to_image(structure, absfilename, fmt="gif")

    elif fmt in [
        "cif",
        "mcif",
        "poscar",
        "cssr",
        "xsf",
        "mcsqs",
        "yaml",
        "fleur-inpgen",
        "prismatic",
        "res",
    ]:
        structure[-1].to(filename=absfilename, fmt=fmt)  # type: ignore

    else:
        try:
            structure[-1].to(filename=absfilename)
        except Exception as e:
            raise NotImplementedError(
                f"formats other than [pdb, xyz, dump, json, as, hzw, png, gif] are handled by pymatgen, while it returns: {e}",
            )


def convert(
    infile,
    si=None,
    ele=None,
    ai=None,
    infmt: Optional[str] = None,
    task: str = "scf",
    outfile: str = "temp.xyz",
    outfmt: Optional[str] = None,
    coords_are_cartesian: bool = True,
):
    """convert from infile to outfile.

    - multi -> single, only keep last step
    - crystal -> molecule, will lose lattice info
    - molecule -> crystal, will add a box of twice the maximum xyz
    - pdb, dump may suffer decimal precision loss

    Parameters
    ----------
    infile:
        - h5/json/as/hzw/cif/poscar/cssr/xsf/mcsqs/prismatic/yaml/fleur-inpgen file path
        - If a folder is given, will read {task}.h5/json files
        - If structures are given, will read multiple structures.

    si: int, list, or str
        - Structure index, starting from 1

            - si=1, read the 1st
            - si=[1,2], read the 1st and 2nd
            - si=':', read all
            - si='-3:', read the last 3
        - If empty, for multi-configuration files, all configurations will be read; for single-configuration files, the latest configuration will be read.
        - This parameter is only valid for h5/json files.

    ele:
        - Element symbol, written as 'H' or ['H','O']
        - If empty, atomic information for all elements will be read.
        - This parameter is only valid for h5/json files.

    ai:
        - Atom index, starting from 1
        - Usage is the same as si
        - If empty, atomic information for all atoms will be read.
        - This parameter is only valid for h5/json files.

    infmt:
        - Input structure file type, e.g., 'h5'. If None, the file extension will determine the format.

    task:
        - Used when datafile is a folder path to locate the internal {task}.h5/json file.
        - Calculation task type, including 'scf', 'relax', 'neb', 'aimd'. Other values will be ignored.

    outfile:
        - Output filename

    outfmt:
        - Output structure file type, e.g., 'xyz'. If None, the file extension will determine the format.

    coords_are_cartesian:
        - Whether to write coordinates in Cartesian form (default: True); otherwise, fractional coordinates will be used.
        - This option is currently only valid for as and json formats.

    Examples
    --------
    >>> from dspawpy.io.structure import convert
    >>> convert('tests/supplement/PtH.as', outfile='tests/doctest_out/PtH.hzw') # doctest: +ELLIPSIS
    ==> .../PtH...hzw

    batch test

    >>> for readable in ['relax.h5', 'system.json', 'aimd.pdb', 'latestStructure.as', 'CuO.hzw', 'POSCAR']:
    ...     for writable in ['pdb', 'xyz', 'dump', 'as', 'hzw', 'POSCAR']:
    ...         convert('tests/supplement/stru/'+readable, outfile=f"tests/doctest_out/{readable.split('.')[0]}.{writable}") # doctest: +ELLIPSIS
    ==> .../relax...pdb
    ==> .../relax...xyz
    ==> .../relax...dump
    ==> .../relax...as
    ==> .../relax...hzw
    ==> .../system...pdb
    ==> .../system...xyz
    ==> .../system...dump
    ==> .../system...as
    ==> .../system...hzw
    ==> .../aimd...pdb
    ==> .../aimd...xyz
    ==> .../aimd...dump
    ==> .../aimd...as
    ==> .../aimd...hzw
    ==> .../latestStructure...pdb
    ==> .../latestStructure...xyz
    ==> .../latestStructure...dump
    ==> .../latestStructure...as
    ==> .../latestStructure...hzw
    ==> .../CuO...pdb
    ==> .../CuO...xyz
    ==> .../CuO...dump
    ==> .../CuO...as
    ==> .../CuO...hzw
    ==> .../POSCAR...pdb
    ==> .../POSCAR...xyz
    ==> .../POSCAR...dump
    ==> .../POSCAR...as
    ==> .../POSCAR...hzw

    """
    write(read(infile, si, ele, ai, infmt, task), outfile, outfmt, coords_are_cartesian)
